## \package insult
## \copyright Copyright (C) 2016 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import random
import os
import re
import json
import six
from six.moves import filter


class WordList(object):
    """!
    A list of insulting words of some kind
    """

    def __init__(self, id, words=[]):
        """!
        \param id       A WordListId (expanded with word_list_id())
        \param words    An iterable containing words to add to the list
        """
        id = word_list_id(id)

        ## Name of the word list
        self.name = id.name
        ## Set of words in the list
        self.words = set(words)
        ## Flags fot the set of words \todo
        ## \note Two word lists can have the same name iff they have different flags
        self.flags = id.flags

    def add_word(self, word):
        """!
        Appends a word to the list
        """
        self.words.add(word)

    def get(self, max_count=1, min_count=None):
        """!
        Retrieves a random subset of words
        \param max_count Maximum number of words to return
        \param min_count Minimum number of words to return,
                         if omitted, returns exactly \p max_count words
        """
        if max_count > len(self.words):
            max_count = len(self.words)
        if min_count is None:
            min_count = max_count
        return random.sample(self.words, random.randint(min_count, max_count))

    def check_flags(self, flags):
        """!
        Checks if the flags match
        \todo
        """
        return True


class WordListId(object):
    """!
    Identifier for a word list
    """

    def __init__(self, name, flags=0):
        self.name = name
        self.flags = flags

    def check(self, word_list):
        """!
        Checks if the word list matches this Id
        \param word_list a WordList object
        """
        return word_list.name == self.name and word_list.check_flags(self.flags)


def word_list_id(*args):
    """!
    Workaround for the lack of overloading
    """
    if len(args) == 1:
        if isinstance(args[0], six.string_types):
            return WordListId(args[0])
        if isinstance(args[0], WordListId):
            return args[0]
        if type(args[0]) is tuple and len(args[0]) == 2:
            return WordListId(*args[0])
    elif len(args) == 2:
        return WordListId(*args)
    raise TypeError("Invalid arguments to the WordListId constructor")


class Insulter(object):
    """!
    Object that can generate insults
    \note all methods that take a WordListId argument, will expand it
          using word_list_id()
    """

    ## Regex used to recognize valid word file names
    regex_word_file = re.compile(r'^[a-z_]+$')
    ## File name for language rules
    rules_file = "rules.json"

    def __init__(self):
        """!
        """
        ## Word lists to look up
        self.word_lists = []
        ## Maximum number of repetitions allowed
        self.max_count = None
        ## Maximum number of repetitions allowed for a specific word list
        self.list_max_count = {}
        ## Language rules
        self.rules = {}

    def word_list(self, wl_id, add=False):
        """!
        Returns a matching word list
        \param wl_id    WordListId to match the list
        \param add      if \c True, missing lists will be added to the insulter
        \throws Exception if the list cannot be retrieved
        """
        wl_id = word_list_id(wl_id)
        for word_list in self.word_lists:
            if wl_id.check(word_list):
                return word_list

        if add:
            self.word_lists.append(WordList(wl_id, []))
            return self.word_lists[-1]

        raise Exception("Word list not found: %s" % wl_id.name)

    def load_directory(self, path):
        """!
        Loads all word lists in \p path
        \param path The path to the directory to load
        \note Only files matching regex_word_file will be considered
        """
        for basename in os.listdir(path):
            full = os.path.join(path, basename)
            if basename == Insulter.rules_file:
                try:
                    with open(full) as file:
                        rules = json.load(file)
                    for name, rule in six.iteritems(rules):
                        self.set_rules(name, rule)
                except ValueError:
                    pass
            elif os.path.isfile(full) and self.regex_word_file.match(basename):
                with open(full) as file:
                    lines = list(filter(bool, (line.strip() for line in file)))
                self.add_words(basename, lines)

    def add_words(self, wl_id, words):
        """!
        Adds words to a word list
        \param wl_id    WordListId to match the list
        \param words    Iterable with words to be added
        """
        self.word_list(wl_id, True).words |= set(words)

    def get(self, wl_id, max_count=1, min_count=None):
        """!
        Retrieves a random subset of words form a word list
        \param wl_id     WordListId to match the list
        \param max_count Maximum number of words to return,
                         note that it will be checked against the value set
                         with set_max()
        \param min_count Minimum number of words to return,
                         if omitted, returns exactly \p max_count words
        \see set_max(), WordList.get()
        """
        wl_id = word_list_id(wl_id)
        if self.max_count is not None and max_count > self.max_count:
            max_count = self.max_count
        if max_count > self.list_max_count.get(wl_id.name, max_count):
            max_count = self.list_max_count[wl_id.name]
        return self.word_list(wl_id).get(max_count, min_count)

    def format(self, string):
        """!
        Formats an insult string
        """
        doc = NotQuiteXml(string, lambda e: e.attrs.get("_expansion", ""))

        new_unexpanded = [
            elem
            for elem in doc.contents
            if isinstance(elem, NotQuiteXmlElement)
        ]
        unexpanded = []

        while len(new_unexpanded) != len(unexpanded) and new_unexpanded:
            unexpanded = new_unexpanded
            new_unexpanded = []
            for elem in unexpanded:
                if not self._expand_element(elem, doc):
                    new_unexpanded.append(elem)
        return str(doc)

    def _expand_element(self, element, doc):
        try:
            if element.tag_name in self.rules:
                return self._expand_rule(element, doc)
            min = None
            max = 1
            if "count" in element.attrs:
                max = int(element.count)
            elif "max" in element.attrs:
                max = int(element.max)
                min = int(element.attrs.get("min", element.max))
            element._expansion = " ".join(self.get(element.tag_name, max, min))
            return True
        except Exception:
            return False

    def _expand_rule(self, element, doc):
        for rule in self.rules[element.tag_name]:
            pattern = rule["target"]
            target = doc.element_by_id(element.target)
            if re.match(pattern, target._expansion):
                element._expansion = re.sub(pattern, rule["result"], target._expansion)
                return True
        return False

    def set_max(self, max_count, word_list=None):
        """!
        Set the maximum number of repetitions for get()
        \param max_count    Maximum to be set,
                            if \c None will disable the maximum limit
        \param word_list    Word list name (note: not a WordListId)
        """
        if word_list is None:
            self.max_count = max_count
        elif max_count is None:
            if str(word_list) in self.list_max_count:
                del self.list_max_count[str(word_list)]
        else:
            self.list_max_count[str(word_list)] = max_count

    def set_rules(self, name, rules):
        """!
        Sets some language rules for the given identifier
        """
        self.rules[name] = rules


class NotQuiteXml(object):
    """!
    Parses strings with flat xml elements intermixed with text
    Only a very minimal subset of xml/sgml is supported
    """
    _entities = {
        "lt": "<",
        "gt": ">",
        "amp": "&",
    }

    def __init__(self, contents=None, to_string=lambda x: ""):
        """!
        \param contents     \c None or a sting to be parsed
        \param to_string    A functor to convert NotQuiteXmlElement
                            objects to a string
        """
        ## List mixing strings and NotQuiteXmlElement elements
        self.contents = []
        ## Elements with an ID
        self.elements_with_id = {}
        ## Functor to convert NotQuiteXmlElement objects to a string
        self.to_string = to_string

        if isinstance(contents, six.string_types):
            self.parse_string(contents)

    def element_by_id(self, id):
        """!
        Returns the element matching the given id
        """
        return self.elements_with_id[id]

    def elements_by_tag_name(self, name):
        """!
        Returns a list of elements with the given tag name
        """
        return [
            element
            for element in self.contents
            if isinstance(element, NotQuiteXmlElement) and
            element.tag_name == name
        ]

    def elements_by_attribute(self, name, value):
        """!
        Returns a list of elements having the given attribute with the given value
        """
        return [
            element
            for element in self.contents
            if isinstance(element, NotQuiteXmlElement) and
            name in element.attrs and element.attrs[name] == value
        ]

    def __str__(self):
        """!
        Converts the document to a string (using self.to_string for elements)
        """
        return "".join(str(elem) for elem in self.contents)

    def __repr__(self):
        return "".join(
            elem if type(elem) is str else repr(elem)
            for elem in self.contents
        )

    def parse_string(self, string):
        """!
        Parses a string into self.contents
        """
        self.contents = list(self._lex_text(iter(string)))

    def _lex_text(self, iterator):
        """!
        Internal lexer, starting state (text)
        yields elements for self.contents
        \param iterator A character iterator
        """
        string = ""
        try:
            while True:
                ch = next(iterator)
                if ch == "<":
                    if string:
                        yield string
                    string = ""
                    yield self._lex_elem_name(iterator)
                elif ch == "&":
                    string += self._lex_entity(iterator)
                else:
                    string += ch
        except StopIteration:
            if string:
                yield string

    def _lex_entity(self, iterator):
        """!
        Internal lexer, entity state
        in: (text) -> &
        out: ; -> (text)
        \returns A string corresponding to the entity
        \param iterator A character iterator
        """
        name = ""
        while True:
            ch = next(iterator)
            if ch == ";":
                break
            else:
                name += ch
        return self._entities.get(name, "")

    def _lex_elem_name(self, iterator):
        """!
        Internal lexer, element name state
        in: (text) -> <
        out: /> | > | _ -> (attrs) -> (text)
        \returns A string corresponding to the entity
        \param iterator A character iterator
        """
        name = ""
        while True:
            ch = next(iterator)
            if ch.isspace() or ch in "/>":
                element = NotQuiteXmlElement(self, name)
                self._lex_elem_attrs(ch, iterator, element)
                return element
            name += ch

    def _lex_elem_attrs(self, och, iterator, element):
        """!
        Internal lexer, element attributes state
        in: (text) -> (element name) -> _ | > | />
        out: /> | > | -> (text)
        \param och      Character used to enter this state
        \param iterator A character iterator
        \param element  Element to set the attributes to
        """
        ch = och
        while True:
            if ch == ">":
                return element
            elif ch == "/":
                next(iterator)
                return element
            elif ch.isspace() or ch == "":
                ch = next(iterator)
            else:
                ch = self._lex_elem_attr_name(ch, iterator, element)

    def _lex_elem_attr_name(self, och, iterator, element):
        """!
        Internal lexer, element attribute name
        in: (attrs) -> [not space or tag end]
        out: = -> (attr_value) -> (attrs)
        out: /> | > | _ -> (attrs)
        \param och      Character used to enter this state
        \param iterator A character iterator
        \param element  Element to set the attributes to
        \returns A lookahead character
        """
        name = och
        while True:
            ch = next(iterator)
            if ch.isspace() or ch in "/>":
                if name != "id":
                    element[name] = name
                return ch
            elif ch == "=":
                return self._lex_elem_attr_value(name, iterator, element)
            name += ch

    def _lex_elem_attr_value(self, name, iterator, element):
        """!
        Internal lexer, element attribute value
        in: (attr_name) -> =
        out: (attr_name) -> =" -> (here) -> " -> (attrs)
        out: (attr_name) -> =' -> (here) -> ' -> (attrs)
        out: (attr_name) -> = -> (here) -> _ -> (attrs)
        \param name     Name of the attribute
        \param iterator A character iterator
        \param element  Element to set the attributes to
        \returns A lookahead character
        """
        ch = next(iterator)
        value = ""
        skip = True
        if ch == '"' or ch == "'":
            delim = ch
            break_cond = lambda char: char == delim
        elif ch.isspace() or ch in "/>":
            break_cond = lambda char: True
            skip = False
        else:
            break_cond = lambda char: char.isspace() or char in "/>"
            value = ch
            skip = False

        ch = next(iterator)
        while not break_cond(ch):
            value += ch
            ch = next(iterator)

        self._elem_set_attr(element, name, value)
        if skip:
            ch = next(iterator)
        return ch

    def _elem_set_attr(self, element, name, value):
        """!
        Internal, sets an element attribute during parsing
        \todo handle this where you can set attributes in NotQuiteXmlElement
        """
        if name == "id":
            if value not in self.elements_with_id:
                element.id = value if value else None
        else:
            element[name] = value


class NotQuiteXmlElement(object):
    """!
    Non-text element in a NotQuiteXml document.
    Attributes can be accessed with the subscript operator or as members
    (if they don't clash with other members)
    """
    def __init__(self, document, tag_name, id=None, attrs={}):
        """!
        \param document     A NotQuiteXml object which contains this element
        \param tag_name     Name of the element tag in the source string
        \param id           Element id, must be unique in \p document or None
        \param attrs        Extra attribues
        """
        self._document = document
        self._tag_name = tag_name
        self.attrs = attrs.copy()
        self._id = None # this must be last (see __setattr__)
        self.id = id

    @property
    def document(self):
        return self._document

    @property
    def tag_name(self):
        return self._tag_name

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        if value == self._id:
            return

        if value is None:
            del self.id
            return

        if value in self.document.elements_with_id:
            raise KeyError("ID already in use: %s" % value)

        if self._id is not None:
            del self.document.elements_with_id[self._id]
        self._id = value
        self.document.elements_with_id[value] = self

    @id.deleter
    def id(self):
        if self._id is not None:
            del self.document.elements_with_id[self._id]
            self._id = None

    def __getitem__(self, key):
        return self.attrs[key]

    def __setitem__(self, key, value):
        self.attrs[key] = value

    def __delitem__(self, key):
        del self.attrs[key]

    def __getattr__(self, name):
        try:
            super(NotQuiteXmlElement, self).__getattr__(name)
        except AttributeError:
            if name in self.attrs:
                return self.attrs[name]
            else:
                raise

    def __setattr__(self, name, value):
        if name in dir(self) or "_id" not in dir(self):
            super(NotQuiteXmlElement, self).__setattr__(name, value)
        else:
            self.attrs[name] = value

    def __delattr__(self, name):
        try:
            super(NotQuiteXmlElement, self).__delattr__(name)
        except AttributeError:
            if name in self.attrs:
                del self.attrs[name]
            else:
                raise

    def __str__(self):
        """!
        Converts the element to a string using the document to_string attribute
        """
        return self.document.to_string(self)

    def __repr__(self):
        return "<%s%s%s/>" % (
            self.tag_name,
            " id=\"%s\"" % self._id if self._id is not None else "",
            "".join(" %s=\"%s\"" % attr for attr in six.iteritems(self.attrs))
        )



#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import unittest
import mock
from io import StringIO
import os
import sys

sys.path.append(os.path.join(os.path.dirname(os.path.abspath(os.path.dirname(__file__))), "lib"))
import insult

class TestWordList(unittest.TestCase):
    def test_ctor(self):
        wl = insult.WordList("name")
        self.assertEqual(wl.name, "name")
        self.assertFalse(wl.words)

        wl = insult.WordList("name", ["hello", "world"])
        self.assertEqual(wl.name, "name")
        self.assertIn("hello", wl.words)
        self.assertIn("world", wl.words)

    def test_add_word(self):
        wl = insult.WordList("name")
        self.assertEqual(wl.name, "name")
        self.assertNotIn("hello", wl.words)
        wl.add_word("hello")
        self.assertIn("hello", wl.words)

    def test_get(self):
        wl = insult.WordList("name", ["a", "b", "c", "d"])
        self.assertEqual(len(wl.get()), 1)
        self.assertIn(wl.get()[0], wl.words)
        self.assertEqual(len(wl.get(3)), 3)
        self.assertEqual(len(wl.get(10)), 4)
        self.assertLessEqual(len(wl.get(3, 2)), 3)
        self.assertGreaterEqual(len(wl.get(3, 2)), 2)

    def test_word_list_id_class(self):
        wl1 = insult.WordList("1")
        wl2 = insult.WordList("2")
        id = insult.WordListId("1")
        self.assertTrue(id.check(wl1))
        self.assertFalse(id.check(wl2))

    def test_word_list_id_def(self):
        wli = insult.word_list_id("hello")
        self.assertIsInstance(wli, insult.WordListId)
        self.assertEqual(wli.name, "hello")
        self.assertEqual(wli.flags, 0)

        wli = insult.word_list_id("hello", 42)
        self.assertIsInstance(wli, insult.WordListId)
        self.assertEqual(wli.name, "hello")
        self.assertEqual(wli.flags, 42)

        wli = insult.word_list_id(("hello", 42))
        self.assertIsInstance(wli, insult.WordListId)
        self.assertEqual(wli.name, "hello")
        self.assertEqual(wli.flags, 42)

        wli = insult.word_list_id(insult.WordListId("hello", 42))
        self.assertIsInstance(wli, insult.WordListId)
        self.assertEqual(wli.name, "hello")
        self.assertEqual(wli.flags, 42)

        self.assertRaises(TypeError, lambda: insult.word_list_id(1, 2, 3, 4))
        self.assertRaises(TypeError, lambda: insult.word_list_id((1, 2, 3, 4)))


class TestInsulter(unittest.TestCase):
    def test_ctor(self):
        ins = insult.Insulter()
        self.assertFalse(ins.word_lists)
        self.assertIsNone(ins.max_count)
        self.assertFalse(ins.list_max_count)

    def test_word_list(self):
        ins = insult.Insulter()
        self.assertRaises(Exception, lambda: ins.word_list("a"))
        self.assertIsInstance(ins.word_list("a", True), insult.WordList)
        self.assertEqual(ins.word_list("a").name, "a")

    @mock.patch("os.listdir", return_value=["foo", "bar", "lol.jpg", insult.Insulter.rules_file])
    @mock.patch("os.path.isfile", return_value=True)
    def test_load_directory(self, *args, **kwargs):
        def fake_open(file, mode=""):
            if file == "foo/foo":
                contents = "foo\nbar\n"
            elif file == "foo/bar":
                contents = "hello\nworld\n\n"
            elif file == "foo/rules.json":
                contents = '{"test":[{"hello":"world"}]}'
            elif file == "bar/rules.json":
                contents = 'invalid json'
            elif file.startswith("bar/"):
                contents = ""
            else:
                raise Exception("Wrong file open: %s" % file)
            io = StringIO(contents)
            io.__enter__ = lambda *a, **k: io
            io.__exit__ = lambda *a, **k: None
            return io

        with mock.patch("insult.open", side_effect=fake_open):
            ins = insult.Insulter()
            ins.load_directory("foo")
            self.assertSetEqual(set(["foo", "bar"]), ins.word_list("foo").words)
            self.assertSetEqual(set(["hello", "world"]), ins.word_list("bar").words)
            self.assertRaises(Exception, lambda: ins.word_list("lol.jpg"))
            self.assertEqual(ins.rules["test"], [{"hello": "world"}])

            ins = insult.Insulter()
            try:
                ins.load_directory("bar")
            except Exception:
                self.fail("load_directory shouldn't fail")

    def test_add_words(self):
        ins = insult.Insulter()
        ins.word_list("a", True)
        ins.add_words("a", ["hello"])
        self.assertIn("hello", ins.word_list("a").words)

        ins.add_words("b", ["world"])
        self.assertIn("world", ins.word_list("b").words)

    def test_get(self):
        ins = insult.Insulter()
        ins.add_words("name", ["a", "b", "c", "d"])
        self.assertEqual(len(ins.get("name")), 1)
        self.assertEqual(len(ins.get("name", 3)), 3)
        self.assertEqual(len(ins.get("name", 10)), 4)
        self.assertLessEqual(len(ins.get("name", 3, 2)), 3)
        self.assertGreaterEqual(len(ins.get("name", 3, 2)), 2)

    def test_set_max(self):
        ins = insult.Insulter()
        ins.add_words("name", ["a", "b", "c", "d", "e"])
        self.assertEqual(len(ins.get("name", 5)), 5)

        ins.set_max(4)
        self.assertEqual(len(ins.get("name", 5)), 4)

        ins.set_max(3, "name")
        self.assertEqual(len(ins.get("name", 5)), 3)

        ins.set_max(2)
        self.assertEqual(len(ins.get("name", 5)), 2)

        ins.set_max(None)
        self.assertEqual(len(ins.get("name", 5)), 3)

        ins.set_max(None, "name")
        self.assertEqual(len(ins.get("name", 5)), 5)

        self.assertIsNone(ins.set_max(None, "name"))

    def test_format(self):
        ins = insult.Insulter()
        ins.set_rules("article", [
            {"target": "^[aeiouAEIOU8].*",  "result": r"an "},
            {"target": ".*",                "result": r"a "}
        ])
        phrase = "You are <article target='adj1' /><adjective id='adj1' count='3' /> <animal />"

        ins.word_lists = []
        ins.add_words("animal", ["pig"])
        ins.add_words("adjective", ["ugly"])
        self.assertEqual(ins.format(phrase), "You are an ugly pig")

        ins.word_lists = []
        ins.add_words("animal", ["pig"])
        ins.add_words("adjective", ["ignorant", "ugly", "awful"])
        formatted = ins.format(phrase)
        self.assertTrue(formatted.startswith("You are an "))
        self.assertIn("ignorant", formatted)
        self.assertIn("ugly", formatted)
        self.assertIn("awful", formatted)
        self.assertTrue(formatted.endswith("pig"))

        ins.word_lists = []
        ins.add_words("animal", ["pig"])
        ins.add_words("adjective", ["dastardly", "grotesque", "smelly"])
        formatted = ins.format(phrase)
        self.assertTrue(formatted.startswith("You are a "))
        self.assertIn("dastardly", formatted)
        self.assertIn("grotesque", formatted)
        self.assertIn("smelly", formatted)
        self.assertTrue(formatted.endswith("pig"))

        phrase = "You are <article target='adj1' /><adjective id='adj1' max='3' /> <animal />"
        ins.word_lists = []
        ins.add_words("animal", ["pig"])
        ins.add_words("adjective", ["ignorant", "ugly", "awful"])
        formatted = ins.format(phrase)
        self.assertTrue(formatted.startswith("You are an "))
        self.assertIn("ignorant", formatted)
        self.assertIn("ugly", formatted)
        self.assertIn("awful", formatted)
        self.assertTrue(formatted.endswith("pig"))

        self.assertEqual(ins.format("You are <foobar/>here"), "You are here")
        infref = "You are<article target='a' id='b'/> <article target='b' id='a'/>here"
        self.assertEqual(ins.format(infref), "You are here")
        ins.set_rules("article", [])

        ins.word_lists = []
        ins.add_words("animal", ["pig"])
        ins.add_words("adjective", ["smelly"])
        ins.set_rules("article", [])
        self.assertEqual(ins.format("<article target='a'><animal id='a'>"), "pig")


class TestNotQuiteXml(unittest.TestCase):
    def test_ctor(self):
        nqx = insult.NotQuiteXml()
        self.assertFalse(nqx.contents)

    def test_parse_just_text(self):
        nqx = insult.NotQuiteXml("Hello world")
        self.assertEqual(nqx.contents, ["Hello world"])
        self.assertFalse(nqx.elements_with_id)
        self.assertEqual(str(nqx), "Hello world")

    def test_parse_entities(self):
        nqx = insult.NotQuiteXml("Hello &lt;world&gt; &amp;amp;")
        self.assertEqual(nqx.contents, ["Hello <world> &amp;"])

    def test_parse_single_element(self):
        nqx = insult.NotQuiteXml("<hello/>")
        self.assertEqual(len(nqx.contents), 1)
        self.assertIsInstance(nqx.contents[0], insult.NotQuiteXmlElement)
        self.assertIs(nqx.contents[0].document, nqx)
        self.assertEqual(nqx.contents[0].tag_name, "hello")
        self.assertIsNone(nqx.contents[0].id)
        self.assertFalse(nqx.contents[0].attrs)

        nqx = insult.NotQuiteXml("<hello>")
        self.assertEqual(len(nqx.contents), 1)
        self.assertIsInstance(nqx.contents[0], insult.NotQuiteXmlElement)
        self.assertIs(nqx.contents[0].document, nqx)
        self.assertEqual(nqx.contents[0].tag_name, "hello")
        self.assertIsNone(nqx.contents[0].id)
        self.assertFalse(nqx.contents[0].attrs)

        nqx = insult.NotQuiteXml("<hello foo='bar'/>")
        self.assertEqual(len(nqx.contents), 1)
        self.assertIsInstance(nqx.contents[0], insult.NotQuiteXmlElement)
        self.assertIs(nqx.contents[0].document, nqx)
        self.assertEqual(nqx.contents[0].tag_name, "hello")
        self.assertIsNone(nqx.contents[0].id)
        self.assertEqual(nqx.contents[0].attrs["foo"], "bar")

        nqx = insult.NotQuiteXml("<hello foo=bar/>")
        self.assertEqual(len(nqx.contents), 1)
        self.assertIsInstance(nqx.contents[0], insult.NotQuiteXmlElement)
        self.assertIs(nqx.contents[0].document, nqx)
        self.assertEqual(nqx.contents[0].tag_name, "hello")
        self.assertIsNone(nqx.contents[0].id)
        self.assertEqual(nqx.contents[0].attrs["foo"], "bar")

        nqx = insult.NotQuiteXml("<hello foo='bar'>")
        self.assertEqual(len(nqx.contents), 1)
        self.assertIsInstance(nqx.contents[0], insult.NotQuiteXmlElement)
        self.assertIs(nqx.contents[0].document, nqx)
        self.assertEqual(nqx.contents[0].tag_name, "hello")
        self.assertIsNone(nqx.contents[0].id)
        self.assertEqual(nqx.contents[0].attrs["foo"], "bar")

        nqx = insult.NotQuiteXml("<hello foo='bar'   bar=\"foo\" foobar hello= />")
        self.assertEqual(len(nqx.contents), 1)
        self.assertIsInstance(nqx.contents[0], insult.NotQuiteXmlElement)
        self.assertIs(nqx.contents[0].document, nqx)
        self.assertEqual(nqx.contents[0].tag_name, "hello")
        self.assertIsNone(nqx.contents[0].id)
        self.assertEqual(nqx.contents[0].attrs["foo"], "bar")
        self.assertEqual(nqx.contents[0].attrs["bar"], "foo")
        self.assertEqual(nqx.contents[0].attrs["foobar"], "foobar")
        self.assertEqual(nqx.contents[0].attrs["hello"], "")

        nqx = insult.NotQuiteXml("<hello foo='bar' id='hello1'/>")
        self.assertEqual(len(nqx.contents), 1)
        self.assertIsInstance(nqx.contents[0], insult.NotQuiteXmlElement)
        self.assertIs(nqx.contents[0].document, nqx)
        self.assertEqual(nqx.contents[0].tag_name, "hello")
        self.assertEqual(nqx.contents[0].id, "hello1")
        self.assertEqual(nqx.contents[0].attrs["foo"], "bar")
        self.assertNotIn("id", nqx.contents[0].attrs)

        nqx = insult.NotQuiteXml("<hello id/>")
        self.assertEqual(len(nqx.contents), 1)
        self.assertIsInstance(nqx.contents[0], insult.NotQuiteXmlElement)
        self.assertIs(nqx.contents[0].document, nqx)
        self.assertEqual(nqx.contents[0].tag_name, "hello")
        self.assertFalse(nqx.contents[0].attrs)
        self.assertNotIn("id", nqx.contents[0].attrs)
        self.assertIsNone(nqx.contents[0].id)

    def test_parse_mixed(self):
        nqx = insult.NotQuiteXml("Hello <world><foo/>bar")
        self.assertEqual(len(nqx.contents), 4)
        self.assertEqual(str(nqx), "Hello bar")
        self.assertIsInstance(nqx.contents[0], str)
        self.assertIsInstance(nqx.contents[1], insult.NotQuiteXmlElement)
        self.assertIsInstance(nqx.contents[2], insult.NotQuiteXmlElement)
        self.assertIsInstance(nqx.contents[3], str)

        nqx = insult.NotQuiteXml("<foo bar=1><foo bar=2>")
        self.assertEqual(len(nqx.contents), 2)
        self.assertIsInstance(nqx.contents[0], insult.NotQuiteXmlElement)
        self.assertIsInstance(nqx.contents[1], insult.NotQuiteXmlElement)
        self.assertEqual(nqx.contents[0].attrs["bar"], "1")
        self.assertEqual(nqx.contents[1].attrs["bar"], "2")

    def test_element_by(self):
        nqx = insult.NotQuiteXml("<foo bar=1><foo id=1 bar=2><bar bar=2 foo=''><bar id=1>")
        self.assertEqual(nqx.element_by_id("1").bar, "2")
        self.assertEqual(nqx.element_by_id("1").tag_name, "foo")
        self.assertEqual(len(nqx.elements_by_tag_name("foo")), 2)
        self.assertEqual(len(nqx.elements_by_attribute("bar", "2")), 2)
        self.assertEqual(len(nqx.elements_by_attribute("foo", "")), 1)

    def test_repr(self):
        nqx = insult.NotQuiteXml("hello <foo><foo bar=1>")
        self.assertEqual(repr(nqx), 'hello <foo/><foo bar=\"1\"/>')

    def test_element_document(self):
        nqx = insult.NotQuiteXml("<hello foo='bar' id='hello1'/>")
        element = nqx.contents[0]

        self.assertIs(element.document, nqx)
        def foo(): element.document = 0
        self.assertRaises(AttributeError, foo)
        self.assertIs(element.document, nqx)

    def test_element_tag_name(self):
        nqx = insult.NotQuiteXml("<hello foo='bar' id='hello1'/>")
        element = nqx.contents[0]

        self.assertEqual(element.tag_name, "hello")
        def foo(): element.tag_name = 0
        self.assertRaises(AttributeError, foo)
        self.assertEqual(element.tag_name, "hello")

    def test_element_id(self):
        nqx = insult.NotQuiteXml("<hello foo='bar' id='hello1'/><world/>")
        element = nqx.contents[0]

        self.assertEqual(element.id, "hello1")
        self.assertIs(nqx.elements_with_id["hello1"], element)
        element.id = "hello2"
        self.assertEqual(element.id, "hello2")
        self.assertNotIn("hello1", nqx.elements_with_id)
        self.assertIs(nqx.elements_with_id["hello2"], element)

        try:
            element.id = "hello2"
        except KeyError:
            self.fail("Should not fail on same id assignment")

        def foo(): nqx.contents[1].id = "hello2"
        self.assertRaises(KeyError, foo)

        element.id = None
        self.assertNotIn("hello2", nqx.elements_with_id)
        element.id = "hello1"
        self.assertIn("hello1", nqx.elements_with_id)
        del element.id
        self.assertNotIn("hello1", nqx.elements_with_id)
        del element.id

    def test_element_attrs(self):
        nqx = insult.NotQuiteXml("<hello foo='bar' id='hello1'/>")
        element = nqx.contents[0]

        self.assertEqual(element.foo, "bar")
        del element.foo
        self.assertNotIn("foo", element.attrs)
        element.foo = "baz"
        self.assertEqual(element.attrs["foo"], "baz")
        def foo(): del element.notattr
        self.assertRaises(AttributeError, foo)

        element.attrs["foo"] = "bar"
        self.assertEqual(element["foo"], "bar")
        del element["foo"]
        self.assertNotIn("foo", element.attrs)
        element["foo"] = "baz"
        self.assertEqual(element.attrs["foo"], "baz")

        element.attrs["foo"] = "bar"
        self.assertEqual(element.attrs["foo"], "bar")
        del element.attrs["foo"]
        self.assertNotIn("foo", element.attrs)
        element.attrs["foo"] = "baz"
        self.assertEqual(element.attrs["foo"], "baz")

    def test_to_string(self):
        nqx = insult.NotQuiteXml("Hello <world> <foo/>bar", lambda x: x.tag_name)
        self.assertEqual(str(nqx), "Hello world foobar")



unittest.main()



#!/usr/bin/env python
#
# Copyright (C) 2016 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from distutils.core import setup

setup(
    name="insult",
    version="0.2",
    description="Python bindings to LibInsult",
    #long_description="",
    author="Mattia Basaglia",
    author_email="mattia.basaglia@gmail.com",
    url="http://mattbas.org/insult",
    package_dir={'': 'lib'},
    py_modules=["insult"],
    scripts=["bin/insult"],
    license="GPLv3+",
    platforms=["any"],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: English",
        "Topic :: Communications :: Chat",
        "Topic :: Utilities",
        "Operating System :: OS Independent",
    ],
)
